# encoding: utf-8

class VisitOrder:

	def __init__ (self, mingle):
		self.content = []
		self.mingle = mingle

	def has_more (self):
		return len(self.content)

	def push (self, arg):
		if len(arg) == 0: return
		if self.mingle: self.content.extend(reversed(arg))
		else:
			for e in reversed(arg): self.content.insert(0, e)

	def pop (self):
		try: return self.content.pop(0)
		except: return None

	def show (self):
		return '    [%s]' % ', '.join([e[:7] for e in self.content])

class Row:

	def __init__ (self, heads, history):
		self.heads = heads
		self.history = history

	def if_done (self, name, target):

		# No need to drop down beyond the last element
		if self.previous == target.name: return

		# Binding top and bottom nodes together
		if target.top:
			self.history.at(target.top).bottom = target.bottom
		self.history.at(target.bottom).top = target.top

		# Binding previous and current nodes together
		target.top = self.previous
		self.history.at(self.previous).bottom = name

		# Bumping the row number another time
		self.row += 1
		target.row = self.row

		# This node is now the last
		target.bottom = None

		# Recording current node as the next previous
		self.previous = name

	def if_not_done (self, name, target):

		# No node can appear before any of its children
		children = self.history.skip_if_done(target.child)
		if len(children): return

		# Bind this node with the previous, if any, or…
		if self.previous:
			target.top = self.previous
			self.history.at(self.previous).bottom = name

		# … record this node as the first in the chain
		else: self.first = name

		# Bumping the row number
		self.row += 1
		target.row = self.row

		# Add parents to the order
		self.order.push(self.history.skip_if_done(target.parent))

		# The current node is the next previous
		self.previous = name

		# The current node is done
		target.done = 1

	def unroll (self, mingle, flip):

		# Visit starts with all the heads
		self.order = VisitOrder(mingle)
		if flip: self.heads.reverse()
		self.order.push(self.heads)

		# Reference to previous node, to build the chain
		self.previous = None

		# Starting over the first row
		self.row = -1

		# The first node
		self.first = None

		while self.order.has_more():

			name = self.order.pop()
			target = self.history.at(name)

			# Even if done, a node can drop down in the chain after its
			# last-calling child
			if target.done: self.if_done(name, target)
			else: self.if_not_done(name, target)

		return self.first

def unroll (heads, history, mingle, flip):
	return Row(heads, history).unroll(mingle, flip)





import bintrees

class VisitOrder:

	def __init__ (self):
		self.content = []

	def has_more (self):
		return len(self.content)

	def push (self, arg):
		if len(arg) == 0: return
		self.content.extend(arg)

	def pop (self):
		try: return self.content.pop(0)
		except: return None

	def show (self):
		return '    [%s]' % ', '.join([e[:7] for e in self.content])

class Grid:

	def __init__ (self):
		self.store = {}

	def at (self, index):
		try:
			return self.store[index]
		except:
			#self.store[index] = bintrees.RBTree()
			self.store[index] = bintrees.BinaryTree()
			return self.store[index]

	def add (self, column, row, name):
		t = self.at(column)
		t.insert(row, name)

	def remove (self, column, row):
		t = self.at(column)
		t.remove(row)

	def upper (self, column, row):
		try:
			key, value = self.at(column).prev_item(row)
			return value
		except KeyError: return None

	def lower (self, column, row):
		try:
			key, value = self.at(column).succ_item(row)
			return value
		except KeyError: return None

class Column:

	def __init__ (self, heads, history):

		self.verbose = 0

		self.first = None
		self.width = -1

		self.heads = heads
		self.history = history

	def update_width (self, value):
		self.width = max(self.width, value)

	def find_column_for_head (self, name):

		target = self.history.at(name)

		# Start at the immediate right of previous head
		previous = self.heads[self.heads.index(name) - 1]
		column = self.history.at(previous).column + 1

		while 1:
			self.grid.add(column, target.row, 'MARKER')
			if self.lower_check(target, column) and self.upper_check(target, column):
				self.grid.add(column, target.row, name)
				target.set_column(column)
				self.update_width(column)
				break

			self.grid.remove(column, target.row)
			column += 1
		return

	# This checks whether the target row overlaps with any arrow between
	# the upper node on the column and its parents
	def upper_check (self, target, column):

		upper = self.grid.upper(column, target.row)
		if not upper: return True
		parents = self.history.at(upper).parent
		if len(parents) == 0: return True
		lowest = max([self.history.at(e).row for e in parents])
		return lowest <= target.row

	# This checks whether the row of the following node on column overlaps
	# with any arrow between the target and its parents
	def lower_check (self, target, column):

		lower = self.grid.lower(column, target.row)
		if not lower: return True
		if len(target.parent) == 0: return True
		lowest = max([self.history.at(e).row for e in target.parent])
		return lowest <= self.history.at(lower).row

	def find_column_for_parents (self, name):

		target = self.history.at(name)

		# Parents are processed in row order, from lower to upper
		target.parent.sort(key=lambda e: self.history.at(e).row, reverse=True)

		for parent in [self.history.at(e) for e in target.parent]:

			# If a parent has already a column, just push its border
			if parent.has_column():
				parent.set_border(target.column)
				continue

			column = self.history.select_starting_column(parent.child)
			while 1:
				self.grid.add(column, parent.row, 'MARKER')

				if self.upper_check(parent, column) and self.lower_check(parent, column):
					self.grid.add(column, parent.row, parent.name)
					parent.set_column(column)
					self.update_width(column)
					break

				self.grid.remove(column, parent.row)
				column += 1
		return

	def unroll (self, flip):

		self.width = -1
		self.grid = Grid()

		# The order starts for the named heads
		order = VisitOrder()
		if flip: self.heads.reverse()
		order.push(self.heads)

		while order.has_more():

			name = order.pop()
			target = self.history.at(name)

			# No node is processed more than once
			if target.done: continue

			# If a node is a named head and has not yet a column assigned, it
			# must look for a valid column on its own
			if target.name in self.heads and not target.has_column():
				self.find_column_for_head (name)

			# The node assigns a column to each of its parents, in order,
			# ensuring each starts off on a valid position
			self.find_column_for_parents (name)

			# Parents are added to the order, then the node is done
			order.push(self.history.skip_if_done(target.parent))
			target.done = 1

		return self.width

def unroll (heads, history, flip):
	return Column(heads, history).unroll(flip)




# -*- encoding: utf-8 -*-

class Column:

	def __init__ (self, color, transition, padding):
		self.color = color
		self.transition = transition
		self.padding = padding

class Layout:

	def __init__ (self, size, hflip, vflip):

		self.size = size
		self.hflip = hflip

		self.layout = []
		self.track = {i:set() for i in range(-1, size)}

		self.ltee = '├' if self.hflip else '┤' # \u 251c or 2524
		self.rtee = '┤' if self.hflip else '├' # \u 2524 or 251c

		if vflip:
			self.lrcorner = '┌' if self.hflip else '┐' # \u 250c or 2510
			self.llcorner = '┐' if self.hflip else '┌' # \u 2510 or 250c
		else:
			self.lrcorner = '└' if self.hflip else '┘' # \u 2514 or 2518
			self.llcorner = '┘' if self.hflip else '└' # \u 2518 or 2514

		self.rarrow = '←' if self.hflip else '→' # \u 2192 or 2190
		self.larrow = '→' if self.hflip else '←' # \u 2190 or 2192

	def put_char(self, name, transition, padding):
		column = Column(31 + name % 6, transition, padding)
		self.layout.append(column)

	def compute_even_column(self, index, target):

		if index == target.column:

			if len(target.parent): padding = '│' # \u2502
			else: padding = ' '

			overlap = []
			for e in self.track[index]:
				if e == target.name: continue
				if e in target.parent: continue
				overlap.append(e)

			if len(overlap): transition = '╳' # \u2573
			else: transition = '•' # \u2022

			self.put_char(target.column, transition, padding)
			return

		if index > target.column:

			if target.name in self.track[index]:
				if len(self.track[index]) > 1:
					self.put_char(index, self.ltee, '│') # \u2502
				else:
					self.put_char(index, self.lrcorner, ' ')
				return

			if len(self.track[index]):
				self.put_char(index, '│', '│')
				return

			for jndex in range(index, self.size):
				if target.name in self.track[jndex]:
					self.put_char(jndex, self.rarrow, ' ')
					return

		else:

			if target.name in self.track[index]:
				if len(self.track[index]) > 1:
					self.put_char(index, self.rtee, '│') # \u2502
				else:
					self.put_char(index, self.llcorner, ' ')
				return

			if len(self.track[index]):
				self.put_char(index, '│', '│')
				return

			for jndex in reversed(range(0, index)):
				if target.name in self.track[jndex]:
					self.put_char(jndex, self.larrow, ' ') # \u2500
					return

		if len(self.track[index]):
			self.put_char(index, '│', '│') # \u2502
		else:
			self.put_char(index, ' ', ' ')

		return

	def compute_odd_column(self, index, target):

		if index > target.column:

			if target.name in self.track[index]:
				self.put_char(index, self.rarrow, ' ')
				return

			for jndex in range(index, self.size):
				if target.name in self.track[jndex]:
					self.put_char(jndex, self.rarrow, ' ')
					return

		else:

			if target.name in self.track[index - 1]:
				self.put_char(index - 1, self.larrow, ' ')
				return

			for jndex in reversed(range(0, index - 1)):
				if target.name in self.track[jndex]:
					self.put_char(jndex, self.larrow, ' ')
					return

		self.put_char(index, ' ', ' ')

	def compute_layout (self, target):

		self.layout = []

		if self.size: self.compute_even_column(0, target)

		for i in range(1, self.size):
			self.compute_odd_column(i, target)
			self.compute_even_column(i, target)

		for track in self.track.values():
			track.discard(target.name)

		for name in target.parent:
			self.track[target.column].add(name)

		if self.hflip: self.layout.reverse()
		return self.draw_transition(), self.draw_padding()

	def draw_padding (self):

		padding = ''
		for i in self.layout:
			padding += '\x1b[%dm%s' % (i.color, i.padding)
		return padding

	def draw_transition (self):

		padding = ''
		for i in self.layout:
			if i.transition == '•': padding += '\x1b[m•'
			else: padding += '\x1b[%dm%s' % (i.color, i.transition)
		return padding




# -*- encoding: utf-8 -*-

class Node:

	def __init__ (self):

		self.name = None
		self.parent = []
		self.child = []

		self.message = None
		self.done = 0

		self.column = -1
		self.border = -1
		self.row = -1

		self.top = None    # Previous commit by line
		self.bottom = None # Next commit by line
		self.left = None   # Previous commit by column
		self.right = None  # Next commit by line

	def add_child (self, name):
		if name not in self.child:
			self.child.append(name)

	def has_column (self):
		return self.column >= 0

	def set_column (self, value):
		self.column = value
		self.set_border(value)

	def set_border (self, value):
		self.border = max(self.border, value)

	def get_indent (self):
		return ' ' * 3 * self.column

	def to_oneline(self):
		data = (self.column, self.row, self.get_indent(), self.name[:7])
		return '(%2d, %2d)%s • \x1b[33m%s\x1b[m' % data

	def to_string(self):
		indent = ' ' * 2 * self.column
		str = "%s  Name {%s}" % (indent, self.name)
		for i in self.parent: str += "\n%sParent {%s}" % (indent, i)
		for i in self.child:  str += "\n%s Child {%s}" % (indent, i)
		return str

class NodeDB:

	def __init__ (self):
		self.store = {}
		self.fake = 0

	def stats (self):
		size = len(self.store)
		return size, size - self.fake, self.fake

	def add_node (self, node):
		self.store[node.name] = node

	def at (self, name):
		return self.store[name]

	def clear (self):
		for node in self.store.values():
			node.done = 0

	def drop_missing_refs (self):

		fakes = []
		for node in self.store.values():

			size = len(node.parent)

			if size == 0: continue

			elif size == 1:

				if node.parent[0] not in self.store:
					node.parent.pop(0)

			else:
				for name in node.parent:
					if name not in self.store:
						fake = Node()
						fake.name = name
						fake.message = ['[…]']
						fakes.append(fake)
						self.fake += 1

		for fake in fakes: self.add_node(fake)

	# Due to excessively restricting size limit, some heads may not appear at
	# all in the database. These heads are removed from the list
	def drop_missing_heads (self, heads):
		if not heads: return []
		available = []
		for name in heads:
			if name in self.store:
				available.append(name)
		return available

	def skip_if_done (self, names):
		result = []
		for name in names:
			if not self.store[name].done:
				result.append(name)
		return result

	def split_assigned_from_missing (self, names):
		assigned = []
		missing = []
		for name in names:
			if self.store[name].has_column():
				assigned.append(name)
			else: missing.append(name)
		return assigned, missing

	def select_highest (self, names, column, default):
		result = []
		for name in names:
			target = self.store[name]
			if target.has_column() and target.column <= column: continue
			result.append(target.row)
		if len(result) == 0: return default
		return min(result)

	def select_bounding_box (self, names, column):
		result = []
		for name in names:
			target = self.store[name]
			if target.has_column() and target.column < column: continue
			result.append(target.row)
		return result

	def select_starting_column (self, names):
		selection = []
		for name in names:
			target = self.store[name]
			if target.has_column():
				selection.append(target.column)
		return min(selection)




# encoding: utf-8

from __future__ import print_function
from subprocess import check_output, STDOUT
import re

def _exact_match (one, two):
	return one == two

def _prefix_match (one, two):
	return one in two

def _get_all_heads (heads):
	seen = set()
	f = seen.add
	return [e[0] for e in heads if not (e[0] in seen or f(e[0]))]

def _get_selected_heads (f, heads, order, all_heads):

	seen = set()
	g = seen.add
	result = []

	for name in order:
		for e in heads:
			if f(name, e[1]):
				# heads.remove(e)
				result.append(e[0])

	if all_heads:
		for e in heads:
			result.append(e[0])

	return [e for e in result if not (e in seen or g(e))]

def _load_HEAD (outputTuple):

	cmdlist = 'git show-ref --heads --head'.split()

	try: output = check_output(cmdlist, stderr=STDOUT, encoding='utf-8')
	except:
		print('No HEAD')
		return False

	exp = re.compile(r'^(.*) HEAD$')

	for line in output.split('\n'):

		if len(line) == 0: continue

		token = exp.match(line)
		if not token: continue

		if outputTuple: return [(token.group(1), 'HEAD')]
		return [token.group(1)]

	# HEAD may not exist
	return False

def _load_heads (opt):

	# Looking for heads, i.e. active branches
	cmdlist = ['git', 'show-ref']
	if not opt.remotes: cmdlist.append('--heads')
	if opt.tags: cmdlist.append('--tags')

	# Invoke Git
	try: git_output = check_output(cmdlist, stderr=STDOUT, encoding='utf-8')
	except:
		print('No HEADs')
		return []

	collected = []
	exp = re.compile(r'^(.*) refs\/.*\/(.*)$')

	# Parsing Git response
	for line in git_output.split('\n'):

		# Skipping empty lines (the last one should be empty)
		if len(line) == 0: continue

		# Matching name and name
		name_n_ref = exp.match(line)

		# Broken ref: display message and skip line
		if not name_n_ref:
			print('No match for (%s)' % line)
			continue

		# Save result in order and by name
		collected.append((name_n_ref.group(1), name_n_ref.group(2)))

	return collected

def hunt (opt):

	if len(opt.targets) == 0 and not opt.heads: return _load_HEAD(False)

	wanted = []
	for e in opt.order:
		if e in opt.targets:
			wanted.append(e)
	for e in opt.targets:
		if e not in wanted:
			wanted.append(e)

	collected = [x for x in _load_heads(opt) + _load_HEAD(True) if x]
	return _get_selected_heads(_exact_match if opt.match else _prefix_match, collected, wanted, opt.heads)




# encoding: utf-8
from __future__ import print_function

import sys
import os
import getopt

class Option:

	def __init__ (self):

		self.verbose = 0

		self.heads   = False
		self.tags    = False
		self.remotes = False
		self.mingle  = False
		self.flip    = False
		self.hflip   = False
		self.vflip   = False

		self.order   = []
		self.targets = []

		self.pretty  = False
		self.limit   = False
		self.match   = False

		self.mode    = 0

		self.needColorTrick = False

		version_file = os.path.join(os.path.dirname(__file__), 'VERSION')
		self.version = open(version_file, 'r').read().strip()

	def override (self, other):

		self.verbose |= other.verbose

		self.heads   |= other.heads
		self.tags    |= other.tags
		self.remotes |= other.remotes
		self.mingle  |= other.mingle
		self.flip    |= other.flip
		self.hflip   |= other.hflip
		self.vflip   |= other.vflip

		if other.order: self.order = other.order
		if other.targets: self.targets = other.targets

		if other.pretty: self.pretty = other.pretty

		self.limit   |= other.limit
		self.match   |= other.match

		self.mode    |= other.mode

		return self

def _print_help ():

	print('Usage: %s [options] targets…' % sys.argv[0])
	print()
	print(' -a, --all, --heads : adds all heads to targets')
	print(' -t, --tags         : adds all tags to targets')
	print(' -r, --remotes      : adds all remote branches to targets')
	print()
	print(' -n<N>, --limit <N>  : cuts history to N commits')
	print(' -p<P>, --pretty <P> : uses P as the pretty format for messages')
	print()
	print(' -o<name[,names...]>, --order <name[,names...]> : show all targets in order')
	print()
	print(' --prefix, --prefix-match   : arguments match refnames by prefix')
	print(' -x, --exact, --exact-match : arguments must match refnames exactly')
	print()
	print(' -M, --mingle                          : interlap commit from parallel branches')
	print(' -F, --flip, --flip-heads              : flip heads from top to bottom')
	print(' -H, --horizontal, --flip-horizontally : flip layout from left to right')
	print(' -V, --vertical, --flip-vertically     : flip layout from top to bottom')
	print()
	print(' -f<name>, --file<name> : load preferences from <name> instead of default .githistorian')
	print()
	print(' -m[0,1], --mode[01] : select original bintrees mode [0, default] or new sortedcontainer mode [1, experimental]')

def _print_version (o):
	print("Git-Historian %s © 2014-2019 Ivan Simonini" % o.version)

def _parse(args, sopts, lopts):

	option = Option()
	filename = '.githistorian'

	try:
		optlist, args = getopt.gnu_getopt(args, sopts, lopts)
	except getopt.GetoptError as err:
		_print_help()
		return False, None

	for key, value in optlist:
		if key in ('-h', '--help'):
			_print_help()
			return False, False
		elif key in ('-v', '--verbose'):
			option.verbose = 1
		elif key in ('-a', '--all', '--heads'):
			option.heads = True
		elif key in ('-t', '--tags'):
			option.tags = True
		elif key in ('-r', '--remotes'):
			option.remotes = True
		elif key in ('-n', '--limit'):
			option.limit = int(value)
		elif key in ('-p', '--pretty'):
			option.pretty = value
		elif key in ('-x', '--exact', '--exact-match'):
			option.match = True
		elif key in ('--prefix', '--prefix-match'):
			option.match = False
		elif key in ('-o', '--order'):
			option.order = value.split(',')
		elif key == '--version':
			_print_version(option)
			return False, False
		elif key in ('-f', '--file'):
			filename = value
		elif key in ('-M', '--mingle'):
			option.mingle = True
		elif key in ('-F', '--flip', '--flip-heads'):
			option.flip = True
		elif key in ('-H', '--horizontal', '--flip-horizontally'):
			option.hflip = True
		elif key in ('-V', '--vertical', '--flip-vertically'):
			option.vflip = True
		elif key in ('-m', '--mode'):
			option.mode = int(value)

	option.targets = args

	return option, filename

def parse (inargs):

	sopts = 'atrhvn:p:xo:MFHVm:'
	lopts = ['help', 'verbose', 'version',
			'all', 'heads', 'tags', 'remotes',
			'limit=', 'pretty=',
			'exact', 'exact-match', 'prefix', 'prefix-match',
			'order=',
			'mingle',
			'flip', 'flip-heads',
			'horizontal', 'flip-horizontally',
			'vertical', 'flip-vertically',
			'mode']

	option, filename = _parse(inargs or sys.argv[1:], sopts+'f:', lopts+['file'])
	if not option: return False

	if filename and os.path.exists(filename):
		token = []
		for line in open(filename, 'r').readlines():
			token.append(line.strip())

		doption, dfile = _parse(token, sopts, lopts)
	else: doption = Option()

	return doption.override(option)




# encoding: utf-8
from __future__ import print_function

# Silencing BROKEN PIPE errors
import sys
if sys.platform != 'win32':
    from signal import signal, SIGPIPE, SIG_DFL
    signal(SIGPIPE, SIG_DFL)

from .row import unroll as row_unroll
from .column import unroll as column_unroll
from .layout import Layout

class VisitOrder:

	def __init__ (self):
		self.content = []

	def push (self, arg):
		if len(arg) == 0: return
		for e in reversed(arg):
			self.content.insert(0, e)

	def is_empty (self):
		return len(self.content) == 0

	def has_more (self):
		return len(self.content)

	def pop (self):
		try: return self.content.pop(0)
		except: return None

# Bind commits to their children to construct the reverse graph: each parent
# notifies its children so they can reach it
def _bind_children (heads, db):

	order = VisitOrder()
	order.push(heads)

	while order.has_more():

		name = order.pop()
		commit = db.at(name)

		if commit.done: continue

		for i in commit.parent:
			db.at(i).add_child(name)

		order.push(db.skip_if_done(commit.parent))

		commit.done = 1

# Given a complete grid, compute layout and print it, straight or flipped as
# requested by options
def _print_graph (history, first, width, hflip, vflip):

	t = Layout(width + 1, hflip, vflip)
	name = first
	bigblock = []

	while name:

		node = history.at(name)
		transition, padding = t.compute_layout(node)

		block = ['\x1b[m%s\x1b[m %s' % (transition, node.message[0])]
		for i in node.message[1:]: block.append('\x1b[m%s\x1b[m %s' % (padding, i))

		if vflip: bigblock.append('\n'.join(block))
		else: print('\n'.join(block))

		name = node.bottom

	if vflip:
		bigblock.reverse()
		print('\n'.join(bigblock))

def deploy (opt, roots, history):

	_bind_children(roots, history)
	history.clear()
	first = row_unroll(roots, history, opt.mingle, opt.flip)
	history.clear()
	width = column_unroll(roots, history, opt.flip)

	try: _print_graph(history, first, width, opt.hflip, opt.vflip)
	except BrokenPipeError: pass

	return 0





from .layout import Layout

def bind_children(opt, heads, db):

	order = [e for e in heads]

	while order:
		name = order.pop(0)
		print('Visiting {}'.format(name))
		commit = db.at(name)
		if commit.done: continue

		for i in commit.parent: db.at(i).add_child(name)

		order = db.skip_if_done(commit.parent) + order
		commit.done = 1

def reduce_graphs(opt, heads, db):

	grid = []
	counter = 0
	order = [e for e in heads]
	previous = None

	while order:
		name = order.pop(0)
		print('Reducing {}'.format(name))
		commit = db.at(name)
		if commit.done: continue

		for column in grid:
			if False: pass

		grid.append([name])
		commit.set_column(counter)
		commit.row = counter
		counter += 1

		order = db.skip_if_done(commit.parent) + order
		commit.done = 1

		if previous:
			previous.bottom = commit
			commit.top = previous
			previous.right = commit
			commit.left = previous
			print('{} above {}'.format(previous.name, commit.name))
		previous = commit

	print(grid)

# Given a complete grid, compute layout and print it, straight or flipped as
# requested by options
def _print_graph (history, first, width, hflip, vflip):

	t = Layout(width + 1, hflip, vflip)
	name = first
	bigblock = []

	while name:

		node = history.at(name)
		transition, padding = t.compute_layout(node)

		block = ['\x1b[m%s\x1b[m %s' % (transition, node.message[0])]
		for i in node.message[1:]: block.append('\x1b[m%s\x1b[m %s' % (padding, i))

		if vflip: bigblock.append('\n'.join(block))
		else: print('\n'.join(block))

		name = node.bottom

	if vflip:
		bigblock.reverse()
		print('\n'.join(bigblock))

def deploy(opt, heads, db):

	try:
		print(opt)
		print(heads)
		print(db)

		bind_children(opt, heads, db)
		db.clear()
		print('Children are bound')
		reduce_graphs(opt, heads, db)
		db.clear()
		print('Graph was reduced')

		_print_graph(db, heads[0], 20, opt.hflip, opt.vflip)

	except BrokenPipeError: pass
	return 0





from setuptools import setup, find_packages

with open('VERSION', 'r') as ifd:
    version = ifd.read().strip()

with open('README.md', 'r') as ifd:
	readme = ifd.read()

setup(
	name='githistorian',
	version=version,
	url='https://github.com/drachlyznardh/githistorian',
	author='Ivan Simonini',
	author_email='drachlyznardh@gmail.com',
	description='Graph visualizer for git history',
	long_description=readme,
	long_description_content_type='text/markdown',
	package_dir={'':'src'},
	packages=find_packages(where='src'),
	install_requires=['bintrees'],
	setup_requires=['wheel'],
	entry_points = { 'console_scripts': ['githistorian=githistorian.githistorian:tell_the_story'] },
	package_data={'': ['VERSION']},
	classifiers=[
		"Programming Language :: Python :: 3",
		"License :: OSI Approved :: MIT License",
		"Operating System :: OS Independent",
	],
)



# -*- encoding: utf-8 -*-
from __future__ import print_function
from subprocess import check_output, STDOUT

from .hunter.head import hunt as head_hunt
from .hunter.history import hunt as history_hunt
from .option import parse as parse_cmd_args

def tell_the_story(inargs=None):

	opt = parse_cmd_args(inargs)
	if not opt: return

	try: version = check_output('git --version'.split(), stderr=STDOUT, encoding='utf-8').strip()
	except:
		print('Git is not installed')
		return 1

	try: version = [x for x in version.split(' ')[2].split('.')]
	except:
		print('Unrecognized version %s' % version)
		return 1

	try: check_output('git rev-parse --git-dir'.split(), stderr=STDOUT)
	except:
		print('Not a repo')
		return 1

	# Hunting for history
	if int(version[0]) == 2 and int(version[1]) > 11: opt.needColorTrick = True
	targets = head_hunt(opt)
	roots, history = history_hunt(opt, targets, opt.limit)

	if opt.verbose:
		print('Targets list    %s' % opt.targets)
		print('Targets order   %s' % opt.order)
		print('Targets found   %s' % targets)
		print('Roots displayed %s' % roots)
		lines, commits, omitted = history.stats()
		print('Loaded %d commits, %d omitted' % (commits, omitted))

	# Graph unrolling
	if opt.mode == 1:
		from .alternate import deploy
		return deploy(opt, roots, history)

	from .graph import deploy as deploy_graph
	return deploy_graph(opt, roots, history)




# -*- encoding: utf-8 -*-

from subprocess import check_output

from ..node import Node, NodeDB

# Apply user specified pretty format or the default with no preference is
# supplied
def _select_pretty (value):
	if value: return r'--pretty=%H %P#' + value
	return r'--pretty=%H %P#%C(yellow)%h%C(auto)%d%Creset %s %C(bold red)%ar%Cblue %an'

# Invokes git-log with optional size limit to collect commits, their relation
# with others and the custom messages
def _get_history_dump (opt, heads, limit):

	try:
		cmdlist = ['git', 'log', _select_pretty(opt.pretty)]
		if opt.needColorTrick: cmdlist.append('--color=always')
		if limit: cmdlist.append('-n%d' % limit)
		cmdlist.extend(heads)

		return check_output(cmdlist, encoding='utf-8')
	except:
		print('No history')
		return ''

def hunt (opt, heads, limit):

	history = NodeDB()
	current = None

	if heads == None:
		print('No history')
		return [], history

	for line in _get_history_dump(opt, heads, limit).split('\n'):

		# Skipping empty lines (the last one should be empty)
		if len(line) == 0: continue

		if '#' in line:

			# Store node in map if any, then create new one
			if current: history.add_node(current)
			current = Node()

			# Split line over the sharp character
			token = line.split('#', 1)

			# Store name, parents, message
			names = token[0].split()
			current.name = names[0]
			for i in names[1:]: current.parent.append(i)
			current.message = [token[1]]

		else:
			current.message.append(line)

	# Store the last node
	if current: history.add_node(current)

	# Cleaning database from missing refs
	history.drop_missing_refs()

	# Cleaning root list from missing heads
	roots = history.drop_missing_heads(heads)

	return roots, history

